import json

import trafaret as t


ISO8601_DATETIME_REGEX = r'\d+\-\d+\-\d+T\d+\:\d+\:\d+[Z\-\+\:0-9]{0,6}'


"""
Validate and convert Cover-Rage client request to Python dictionary:
{
    'overall_coverage': float(<overall_coverage>),
    'git_commit': str(<git_commit>),
    'uncovered_lines': {
        str(<file 1>): [int(<line number 1>, int(<line number 2>, ..., int(<line number K>,],
        str(<file 2>): [int(<line number 1>, int(<line number 2>, ..., int(<line number L>,],
        str(<file N>): [int(<line number 1>, int(<line number 2>, ..., int(<line number M>,],
    }
}
"""
cover_rage_client_validator = t.Dict({
    t.Key('overall_coverage') >> 'overall_coverage': t.Float,
    t.Key('git_commit') >> 'git_commit': t.String,
    t.Key('uncovered_lines') >> 'uncovered_lines': t.Mapping(t.String, t.List(t.Int)),
})

# Available Github pull requests actions
# https://developer.github.com/v3/activity/events/types/#pullrequestevent
GITHUB_PULL_REQUEST_ACTIONS = (
    'assigned',
    'unassigned',
    'labeled',
    'unlabeled',
    'opened',
    'edited',
    'closed',
    'reopened',
    'synchronize',
)

# Github pull requests actions that triggers create / update Cover-Rage checks
GITHUB_PULL_REQUEST_TRIGGER_CREATE_UPDATE_ACTIONS = (
    'opened',
    'reopened',
    'edited',
)

# Github pull requests actions that triggers removing Cover-Rage check
GITHUB_PULL_REQUEST_TRIGGER_REMOVE_ACTIONS = (
    'closed'
)

# Regular expression for available Github pull requests actions
github_action_regex = r'|'.join([r'\b{action}\b'.format(action=a) for a in GITHUB_PULL_REQUEST_ACTIONS])

# Github statuses
GITHUB_ERROR_STATUS = 'error'
GITHUB_FAILURE_STATUS = 'failure'
GITHUB_PENDING_STATUS = 'pending'
GITHUB_SUCCESS_STATUS = 'success'
GITHUB_AVAILABLE_STATUSES = (GITHUB_ERROR_STATUS, GITHUB_FAILURE_STATUS, GITHUB_PENDING_STATUS, GITHUB_SUCCESS_STATUS)

# Regular expression for available Github statuses
github_status_regex = r'|'.join([r'\b{action}\b'.format(action=a) for a in GITHUB_AVAILABLE_STATUSES])

"""
Validate and convert Github `PUSH` event payload to Python dictionary:
{
    'head_commit': {
        'sha': str(<sha>),
        'created_at': str(<timestamp>),
    },
}

https://developer.github.com/v3/activity/events/types/#pushevent
"""
github_push_event_validator = t.Dict({
    t.Key('head_commit') >> 'head_commit': t.Dict({
        t.Key('id') >> 'sha': t.String(),
        t.Key('timestamp') >> 'created_at': t.String(regex=ISO8601_DATETIME_REGEX),
    }).ignore_extra('*')
}).ignore_extra('*')

"""
Validate and convert Github `PULL REQUEST` event payload to Python dictionary:
{
    'action': str(assigned|unassigned|labeled|unlabeled|opened|edited|closed|reopened|synchronize),
    'pull_request': {
        'id': int(<id>),
        'number': int(<number>),
        'created_at': str(<created_at>),
        'updated_at': str(<updated_at>),
        'url': str(<url>),
        'head_commit': {
            'sha': str(<sha>),
        },
    },
}

https://developer.github.com/v3/activity/events/types/#pullrequestevent
"""
github_pull_request_event_validator = t.Dict({
    t.Key('action') >> 'action': t.String(regex=github_action_regex),
    t.Key('pull_request') >> 'pull_request': t.Dict({
        t.Key('number') >> 'number': t.Int(),
        t.Key('id') >> 'id': t.Int(),
        t.Key('url') >> 'url': t.URL,
        t.Key('created_at') >> 'created_at': t.String(regex=ISO8601_DATETIME_REGEX),
        t.Key('updated_at') >> 'updated_at': t.String(regex=ISO8601_DATETIME_REGEX),
        t.Key('head') >> 'head_commit': t.Dict({
            t.Key('sha') >> 'sha': t.String,
        }).ignore_extra('*')
    }).ignore_extra('*')
}).ignore_extra('*')

"""
Validate and convert Github `Status` for specific `Ref` payload to Python dictionary:
{
    'status': str(failure|pending|success),
    'sha': str(<sha>),
}

https://developer.github.com/v3/repos/statuses/#get-the-combined-status-for-a-specific-ref
"""
github_status_validator = t.Dict({
    t.Key('state') >> 'status': t.String(regex=github_status_regex),
    t.Key('sha') >> 'sha': t.String,
}).ignore_extra('*')


def cover_rage_client_to_redis(client: dict) -> dict:
    # Additional helper to convert Python dict with data received Cover-Rage client to store it in Redis
    return {
        'overall_coverage': client['overall_coverage'],
        'git_commit': client['git_commit'],
        'uncovered_lines': json.dumps(client['uncovered_lines']),
    }


def github_push_to_redis(gh: dict) -> dict:
    # Additional helper to convert Github `PUSH` to flat dict to store it in Redis
    return {
        'created_at': gh['head_commit']['created_at'],
        'sha': gh['head_commit']['sha'],
    }


def github_pull_request_to_redis(gh: dict) -> dict:
    # Additional helper to convert Github `PULL REQUEST` to flat dict to store it in Redis
    return {
        'action': gh['action'],
        'id': gh['pull_request']['id'],
        'number': gh['pull_request']['number'],
        'created_at': gh['pull_request']['created_at'],
        'updated_at': gh['pull_request']['updated_at'],
        'url': gh['pull_request']['url'],
        'sha': gh['pull_request']['head_commit']['sha'],
    }



"""
Hashes:
-------

repositories:
  key: repo_id
  fields:
    - rage_public_token
    - rage_private_token
    - repo_access_token
    - min_good_coverage
    - coverage
  description: stores main data about repo such as access token, Cover-Rage tokens, other settings

repositories_backward:
  key: rage_public_token
  fields:
    - repo_id
  description: required for backward repo search by Cover-Rage public token

commits_coverage:
  key: commit_id
  fields:
    - overall_coverage
    - uncovered_lines (JSON):
      - file 1: [line number 1, line number 2, line number K]
      - file 2: [line number 1, line number 2, line number L]
      ...
      - file N: [line number 1, line number 2, line number M]
    - created_at
    - updated_at (for commit from pull request)
    - action (for commit from pull request)
    - id of pull request (for commit from pull request)
    - number of pull request (for commit from pull request)
    - url of pull request (for commit from pull request)


Hash Keys:
----------

repo_id:
  parts:
    - prefix
    - kind (`gh` or `bb`)
    - owner
    - repo name
  delimiter: /
  examples:
    - r/gh/john_doe/cool_project
    - r/bb/jane_doe/test_project

commit_id:
  parts:
    - prefix
    - repo_id
    - commit hash
  delimiter: #
  examples:
    - c#r/gh/john_doe/cool_project#1234qwerty5678
    - c#r/bb/jane_doe/test_project#0987sha1234
"""

import json
import logging
# from secrets import token_urlsafe  # TODO: uncomment this in Python 3.6

import aioredis
from aioredis.errors import RedisError

from . import settings
from .secrets import token_urlsafe  # TODO: remove this in Python 3.6


__all__ = [
    'create_repo', 'get_repo', 'update_coverage_for_repo',
    'get_commit', 'update_commit', 'delete_commit', 'update_coverage_for_commit'
]


logger = logging.getLogger('cover_rage')

RAGE_SRV_REPO_PREFIX = 'r'
RAGE_SRV_COVERAGE_PREFIX = 'c'


def dict_to_redis_iterable(value: dict) -> dict:
    """
    Covert Python dictionary to Redis HMSET-command format - http://redis.io/commands/hmset
    :param value: dict
    :raises: AssertationError
    :return: list
    """
    assert isinstance(value, dict)
    result = []
    for k, v in value.items():
        result.extend([k, v])
    return result


def convert_bytes_dict_to_string_dict(value: dict) -> dict:

    def convert(data):
        if isinstance(data, bytes):
            return data.decode()
        elif isinstance(data, dict):
            return dict(map(convert, data.items()))
        elif isinstance(data, (list, tuple)):
            return type(data)(map(convert, data))
        else:
            return data

    return convert(value)


def get_repo_id(repo_kind: str, repo_owner: str, repo_name: str) -> str:
    assert isinstance(repo_kind, str)
    assert isinstance(repo_owner, str)
    assert isinstance(repo_name, str)
    assert repo_kind in ('gh', 'bb')
    assert repo_owner
    assert repo_name
    return '{prefix}/{kind}/{owner}/{name}'.format(
        prefix=RAGE_SRV_REPO_PREFIX,
        kind=repo_kind,
        owner=repo_owner,
        name=repo_name
    )


def split_repo_id(repo_id: str) -> (str, str, str):
    assert isinstance(repo_id, str)
    assert repo_id
    repo_parts = tuple(repo_id.split('/'))
    assert len(repo_parts) == 4
    prefix, repo_kind, repo_owner, repo_name = repo_parts
    assert prefix == RAGE_SRV_REPO_PREFIX
    assert repo_kind in ('gh', 'bb')
    assert repo_owner
    assert repo_name
    return repo_kind, repo_owner, repo_name


def get_commit_id(repo_id: str, commit_hash: str) -> str:
    assert isinstance(repo_id, str)
    assert isinstance(commit_hash, str)
    assert repo_id
    assert commit_hash
    return '{prefix}#{repo_id}#{commit_hash}'.format(
        prefix=RAGE_SRV_COVERAGE_PREFIX,
        repo_id=repo_id,
        commit_hash=commit_hash
    )


def split_commit_id(commit_id: str) -> (str, str):
    assert isinstance(commit_id, str)
    assert commit_id
    commit_parts = tuple(commit_id.split('#'))
    assert len(commit_parts) == 3
    prefix, repo_id, commit_id = commit_parts
    assert prefix == RAGE_SRV_COVERAGE_PREFIX
    assert repo_id
    assert commit_id
    return repo_id, commit_id


async def create_redis_pool(loop: object) -> object:
    """
    Helper to create redis pool
    :param loop: event loop
    :return: Redis pool coroutine
    """
    return await aioredis.create_pool((settings.REDIS_HOST, settings.REDIS_PORT), db=settings.REDIS_DB, loop=loop)


# TODO: cover with tests
async def create_repo(
    pool: object,
    repo_id: str,
    repo_access_token: str,
    min_good_coverage=settings.MIN_GOOD_COVERAGE_PERCENTAGE
) -> (str, str):
    """
    Add data about project to Redis
    :param pool: Redis connection
    :param repo_id: repo id, like `/github/alex/test_project`
    :param repo_access_token: repo access token (Github / Bitbucket token)
    :param min_good_coverage: minimal good coverage percentage (for background color of badge)
    :raises: AssertationError
    :return: tuple with Cover-Rage public and private tokens
    """
    rage_public_token = token_urlsafe(16)
    rage_private_token = token_urlsafe(48)

    async with pool.get() as redis:
        redis_already_has_this_repo = await redis.exists(repo_id)

    assert not redis_already_has_this_repo, 'This repo {} has been already added to cover-rage'.format(repo_id)

    repo = {
        'rage_public_token': rage_public_token,
        'rage_private_token': rage_private_token,
        'repo_access_token': repo_access_token,
        'min_good_coverage': min_good_coverage,
    }

    async with pool.get() as redis:
        await redis.hmset(repo_id, *dict_to_redis_iterable(repo))
        await redis.set(rage_public_token, repo_id)

    return rage_public_token, rage_private_token


# TODO: cover with tests
async def get_repo(pool: object, token=None, repo_id=None) -> (str, dict):
    """
    Get repo from Redis by `repo_id` or `token` (Cover-Rage public token).
    One of `repo_id` or `token` should be set - or AssertationError would be raised.
    :param pool: Redis pool
    :param token: Cover-Rage public token (could be None)
    :param repo_id: Redis repo hash key (could be None)
    :raises: AssertationError
    :return: tuple(`repo_id`, repo data)
    """
    if repo_id is not None:
        key = repo_id
    elif token is not None:
        async with pool.get() as redis:
            key = await redis.get(token)
    else:
        key = None
    assert key
    key = key.decode()
    async with pool.get() as redis:
        repo = await redis.hgetall(key)
        assert repo
        repo = convert_bytes_dict_to_string_dict(repo)
        if 'min_good_coverage' in repo:
            repo['min_good_coverage'] = int(repo['min_good_coverage'])
        if 'coverage' in repo:
            repo['coverage'] = int(repo['coverage'])
        return key, repo


# TODO: cover with tests
async def update_coverage_for_repo(pool: object, repo_id: str, coverage: int) -> bool:
    """
    Update repo data with coverage
    :param pool: Redis pool
    :param repo_id: Redis repo hash key
    :param coverage: coverage percentage
    :return: bool
    """
    try:
        async with pool.get() as redis:
            repo_exists = await redis.exists(repo_id)
            if repo_exists:
                await redis.hmset(repo_id, 'coverage', coverage)
                return True
            else:
                return False
    except RedisError as e:
        message = {
            'parameters': {
                'repo_id': repo_id,
                'coverage': coverage,
            },
            'exception': e
        }
        logging.error(str(message))
        return False


# TODO: cover with tests
async def get_commit(pool: object, repo_id: str, commit_hash: str) -> dict:
    """
    Get commit data from Redis
    :param pool: Redis pool
    :param repo_id: Redis repo hash key
    :param commit_hash: commit hash
    :return: dict
    """
    commit_id = get_commit_id(repo_id, commit_hash)
    try:
        async with pool.get() as redis:
            commit = await redis.hgetall(commit_id)
            assert commit
            commit = convert_bytes_dict_to_string_dict(commit)
            if 'uncovered_lines' in commit:
                commit['uncovered_lines'] = json.loads(commit['uncovered_lines'])
            return commit
    except RedisError as e:
        message = {
            'parameters': {
                'repo_id': repo_id,
                'commit_hash': commit_hash,
            },
            'exception': e
        }
        logging.error(str(message))


# TODO: cover with tests
async def update_commit(pool: object, repo_id: str, commit_hash: str, commit_data: dict) -> bool:
    """
    Create mapping `commit_data` for commit with `commit_hash` in repo `repo_id` and return True.
    If mapping for this commit (in this repo) does not exist - return False.
    :param pool: Redis pool
    :param repo_id: Cover-Rage public token
    :param commit_hash: commit hash
    :param commit_data: commit data
    :return: bool
    """
    commit_id = get_commit_id(repo_id, commit_hash)
    try:
        async with pool.get() as redis:
            await redis.hmset(commit_id, *dict_to_redis_iterable(commit_data))
    except RedisError as e:
        message = {
            'parameters': {
                'repo_id': repo_id,
                'commit_hash': commit_hash,
                'commit_data': commit_data,
            },
            'exception': e
        }
        logging.error(str(message))
        return False
    else:
        return True


# TODO: cover with tests
async def delete_commit(pool: object, repo_id: str, commit_hash: str) -> None:
    """
    Delete mapping for commit with `commit_hash` in repo `repo_id` and return True.
    If mapping for this commit (in this repo) does not exist - return False.
    :param pool: Redis pool
    :param repo_id: Cover-Rage public token
    :param commit_hash: commit hash
    :return: bool
    """
    commit_id = get_commit_id(repo_id, commit_hash)
    try:
        async with pool.get() as redis:
            await redis.delete(commit_id)
    except RedisError as e:
        message = {
            'parameters': {
                'repo_id': repo_id,
                'commit_hash': commit_hash,
            },
            'exception': e
        }
        logging.error(str(message))
        return False
    else:
        return True


# TODO: cover with tests
async def update_coverage_for_commit(pool: object, repo_id: str, commit_hash: str, coverage_data: dict) -> bool:
    """
    Update data for commit with `commit_hash` in repo `repo_id` with coverage data and return True.
    If mapping for this commit (in this repo) does not exist - return False.
    :param pool: Redis pool
    :param repo_id: Cover-Rage public token
    :param commit_hash: commit hash
    :param coverage_data: dict with coverage data
    :return: bool
    """
    commit_id = get_commit_id(repo_id, commit_hash)
    try:
        async with pool.get() as redis:
            await redis.hmset(commit_id, *dict_to_redis_iterable(coverage_data))
    except RedisError as e:
        message = {
            'parameters': {
                'repo_id': repo_id,
                'commit_hash': commit_hash,
                'coverage_data': coverage_data,
            },
            'exception': e
        }
        logging.error(str(message))
        return False
    else:
        return True



#!/usr/bin/env python
import json

import asyncio
import sys
import urllib.parse as urlparse

from aiohttp import ClientSession

from . import settings
from .storage import create_redis_pool, get_repo_id, create_repo


async def create_github_web_hooks(
    loop: object,
    repo_owner: str,
    repo_name: str,
    github_token: str,
    public_token: str
) -> (str, str):
    """
    Helper to create Github web hooks for `push` & `pull request` events
    https://developer.github.com/v3/repos/hooks/#create-a-hook

    :param loop: event loop
    :param repo_owner: Github repo owner
    :param repo_name: Github repo name
    :param github_token: Github access token
    :param public_token: Cover-Rage public token
    :return: tuple with Github web hook urls for `push`, `pull request` events
    """
    github_url = 'https://api.github.com/repos/{repo_owner}/{repo_name}/hooks'.format(
        repo_owner=repo_owner, repo_name=repo_name
    )
    web_hook_urls = []
    params = [
        (settings.SRV_API_GITHUB_PUSH_EVENT_HOOK_URL, 'push'),
        (settings.SRV_API_GITHUB_PULL_REQUEST_EVENT_HOOK_URL, 'pull_request'),
    ]
    for hook_url, event in params:
        rage_server_event_hook_url = urlparse.urlunparse(
            (
                settings.SRV_SCHEME,
                settings.SRV_HOST,
                hook_url.format(public_token=public_token),
                '',
                '',
                ''
            )
        )
        async with ClientSession(loop=loop) as session:
            headers = {
                'Authorization': 'token {token}'.format(token=github_token),
                'Content-Type': 'application/json'
            }
            data = {
                'name': 'web',
                'events': [event],
                'config': {
                    'url': rage_server_event_hook_url,
                    'content_type': 'json'
                }
            }
            async with session.post(github_url, headers=headers, data=json.dumps(data)) as response:
                response_json = await response.json()
                if response.status == 201 and 'url' in response_json:
                    web_hook_urls.append(response_json['url'])
                else:
                    web_hook_urls.append(None)
    return tuple(web_hook_urls)


# TODO: Implement Bitbucket web hooks creation
async def create_bitbucket_web_hooks(
    loop: object,
    repo_owner: str,
    repo_name: str,
    bitbucket_token: str,
    public_token: str
) -> (str, str):
    """
    Helper to create Bitbucket web hooks for `push` & `pull request` events
    :param loop: event loop
    :param repo_owner: Bitbucket repo owner
    :param repo_name: Bitbucket repo name
    :param bitbucket_token: Bitbucket access token
    :param public_token: Cover-Rage public token
    :return: tuple with Bitbucket web hook urls for `push`, `pull request` events
    """
    return None, None


async def add_project_to_redis(loop: object, repo_kind: str, repo_owner: str, repo_name: str, repo_token: str) -> None:

    redis_pool = await create_redis_pool(loop)

    # Generate COver-Rage tokens
    repo_id = get_repo_id(repo_kind, repo_owner, repo_name)
    public_token, private_token = await create_repo(redis_pool, repo_id, repo_token)
    assert public_token
    assert private_token
    print('Success: Cover-Rage tokens were generated.')
    print('\n')

    # Create web hook for Github / Bitbucket
    if repo_kind == 'gh':
        push_web_hook_url, pull_request_web_hook_url = await create_github_web_hooks(
            loop, repo_owner, repo_name, repo_token, public_token
        )
    elif repo_kind == 'bb':
        push_web_hook_url, pull_request_web_hook_url = await create_bitbucket_web_hooks(
            loop, repo_owner, repo_name, repo_token, public_token
        )
    else:
        push_web_hook_url = None
        pull_request_web_hook_url = None
    if push_web_hook_url is not None:
        print('Success: web hook for PUSH event was created.')
        print(push_web_hook_url)
    else:
        print('Error: web hook for PUSH event was not created. Please create this web hook manually.')
    print('\n')
    if pull_request_web_hook_url is not None:
        print('Success: web hook for PULL REQUEST event was created.')
        print(pull_request_web_hook_url)
    else:
        print('Error: web hook for PULL REQUEST event was not created. Please create this web hook manually.')
    print('\n')

    print('CI script: rage_client {server_url} {token} </path/to/git/root> </path/to/coverage.xml>'.format(
        server_url=urlparse.urlunparse(
            (
                settings.SRV_SCHEME,
                settings.SRV_HOST,
                settings.SRV_API_SEND_RESULTS_URL.format(public_token=public_token),
                '',
                '',
                '',
            )
        ),
        token=private_token
    ))

    await redis_pool.clear()
    redis_pool.close()


def __main__():  # pragma: no cover

    if len(sys.argv) != 5:
        print('Usage: {} <repo_kind: gh|bb> <repo_owner> <repo_name> <repo_token>'.format(sys.argv[0]))
        exit(1)

    main_loop = asyncio.get_event_loop()
    main_loop.run_until_complete(
        add_project_to_redis(
            main_loop,
            sys.argv[1],
            sys.argv[2],
            sys.argv[3],
            sys.argv[4],
        )
    )



import asyncio
from concurrent.futures import ThreadPoolExecutor

from aiohttp import web as aiohttp_web

from . import settings
from .api_views import (
    BadgeApiView,
    BitbucketPullRequestEventApiView,
    BitbucketPushEventApiView,
    GithubPullRequestEventApiView,
    GithubPushEventApiView,
    ResultsApiView,
    StatusApiView,
)
from cover_rage_server.storage import create_redis_pool


def get_application():  # pragma: no cover

    async def on_shutdown(application):
        """
        Close Redis pool on shutdown
        """
        await application.redis_pool.clear()

    # Event loop config
    loop = asyncio.get_event_loop()
    loop.set_default_executor(ThreadPoolExecutor(settings.THREAD_POOL_SIZE))

    # Create application
    app = aiohttp_web.Application(loop=loop)

    # Attach Redis poll
    redis_pool = loop.run_until_complete(create_redis_pool(loop))
    app.redis_pool = redis_pool
    app.on_shutdown.append(on_shutdown)

    # Routes
    app.router.add_route('GET',  '/api/badge/{token}/', BadgeApiView)
    app.router.add_route('POST', '/api/hooks/bitbucket/{token}push/', BitbucketPushEventApiView)
    app.router.add_route('POST', '/api/hooks/bitbucket/{token}/pull_request/', BitbucketPullRequestEventApiView)
    app.router.add_route('POST', '/api/hooks/github/{token}/push/', GithubPushEventApiView)
    app.router.add_route('POST', '/api/hooks/github/{token}/pull_request/', GithubPullRequestEventApiView)
    app.router.add_route('POST', '/api/results/{token}/', ResultsApiView)
    app.router.add_route('GET',  '/api/status/{token}/{sha}/', StatusApiView)

    return app


def cli_run(argv):  # pragma: no cover
    app = get_application()
    return app



import re


SVG_BADGE = """<svg xmlns="http://www.w3.org/2000/svg" width="99" height="20">
    <linearGradient id="a" x2="0" y2="100%">
        <stop offset="0" stop-color="#bbb" stop-opacity=".1"/>
        <stop offset="1" stop-opacity=".1"/>
    </linearGradient>
    <rect rx="3" width="99" height="20" fill="#555"/>
    <rect rx="3" x="63" width="36" height="20" fill="{color}"/>
    <path fill="{color}" d="M63 0h4v20h-4z"/>
    <rect rx="3" width="99" height="20" fill="url(#a)"/>
    <g fill="#fff" text-anchor="middle" font-family="DejaVu Sans,Verdana,Geneva,sans-serif" font-size="11">
        <text x="32.5" y="15" fill="#010101" fill-opacity=".3">coverage</text>
        <text x="32.5" y="14">coverage</text>
        <text x="80" y="15" fill="#010101" fill-opacity=".3">{coverage}%</text>
        <text x="80" y="14">{coverage}%</text>
    </g>
</svg>"""


def render_badge(coverage: int, coverage_ok: bool) -> bytes:
    """
    Helper to render SVG coverage badge
    :param coverage: coverage percentage
    :param coverage_ok: good / bad coverage (green / red background, or orange if coverage is undefined (equals 0))
    :return: bytes
    """
    if coverage == 0:
        color = '#BA5106'
    elif coverage_ok:
        color = '#15BA06'
    else:
        color = '#F02E2E'
    return re.sub(r'\n[\s]*', '', SVG_BADGE.format(coverage=coverage, color=color)).encode(encoding='utf-8')



import json
import urllib.parse as urlparse
from copy import deepcopy

import trafaret
from aiohttp import web, ClientSession, MultiDict
from aiohttp.web_exceptions import HTTPUnauthorized, HTTPForbidden, HTTPNotFound

from . import settings
from .badge import render_badge
from .api_validators import (
    cover_rage_client_validator,
    GITHUB_PULL_REQUEST_TRIGGER_CREATE_UPDATE_ACTIONS,
    GITHUB_PULL_REQUEST_TRIGGER_REMOVE_ACTIONS,
    github_push_event_validator,
    github_pull_request_event_validator,
    github_status_validator,
    GITHUB_AVAILABLE_STATUSES,
    GITHUB_ERROR_STATUS,
    GITHUB_FAILURE_STATUS,
    GITHUB_PENDING_STATUS,
    GITHUB_SUCCESS_STATUS,
    cover_rage_client_to_redis,
    github_push_to_redis,
    github_pull_request_to_redis,
)
from .storage import (
    get_repo, update_coverage_for_repo, split_repo_id, get_commit, update_commit, update_coverage_for_commit
)

HTTP_OK = 200
HTTP_ACCEPTED = 202
HTTP_BAD_REQUEST = 400


COVER_RAGE_CONTEXT = 'continuous-integration/cover_rage'


FAILURE_CHECK_DESCRIPTION = 'Not all lines are covered with tests'
PENDING_CHECK_DESCRIPTION = 'Pending coverage check'
SUCCESS_CHECK_DESCRIPTION = 'All new lines are covered with tests'
ERROR_CHECK_CLOSED_PULL_REQUEST_DESCRIPTION = 'Pull request was closed'


def cover_rage_status_url(public_token: str, sha: str) -> str:
    """
    Helper to get status url for `sha` commit
    :param public_token: Cover-Rage app public token
    :param sha: git commit hash
    :return: absolute url for `sha` commit check
    """
    url = urlparse.urlunparse(
        (
            settings.SRV_SCHEME,
            settings.SRV_HOST,
            settings.SRV_STATUS_URL.format(public_token=public_token, sha=sha),
            '',
            '',
            ''
        )
    )
    return url


# TODO: cover with tests
async def send_github_status(
    loop: object,
    repo_owner: str,
    repo_name: str,
    sha: str,
    state: str,
    target_url: str,
    description: str,
    context: str,
    github_access_token: str
) -> bool:
    """
    Helper to set Github status
    :param loop: event loop
    :param repo_owner: repo owner
    :param repo_name: repo name
    :param sha: commit hash
    :param state: status, could be one of: error / failure / pending / success
    :param target_url: url for Cover-Rage check of commit
    :param description: short description of status
    :param context: context identifier
    :param github_access_token: Github access token
    :raises: AssertationError
    :return: True if setting status was successful
    """
    assert isinstance(repo_owner, str)
    assert isinstance(repo_name, str)
    assert isinstance(sha, str)
    assert isinstance(state, str)
    assert isinstance(target_url, str)
    assert isinstance(description, str)
    assert isinstance(context, str)
    assert isinstance(github_access_token, str)
    assert repo_owner
    assert repo_name
    assert sha
    assert state in GITHUB_AVAILABLE_STATUSES
    assert target_url
    assert description
    assert context
    assert github_access_token

    github_url = 'https://api.github.com/repos/{repo_owner}/{repo_name}/statuses/{sha}'.format(
        repo_owner=repo_owner, repo_name=repo_name, sha=sha
    )

    async with ClientSession(loop=loop) as session:
        headers = {
            'Authorization': 'token {token}'.format(token=github_access_token),
            'Content-Type': 'application/json'
        }
        data = {
            'state': state,
            'target_url': target_url,
            'description': description,
            'context': context,
        }
        async with session.post(github_url, headers=headers, data=json.dumps(data)) as response:
            return response.status == 201


# TODO: cover with tests
async def get_github_status(loop: object, repo_owner: str, repo_name: str, sha: str, github_access_token: str) -> str:
    """
    Helper to get Github status
    :param loop: event loop
    :param repo_owner: repo owner
    :param repo_name: repo name
    :param sha: commit hash
    :param github_access_token: Github access token
    :raises: AssertationError
    :return: Github status, could be one of: error / failure / pending / success
    """
    assert isinstance(repo_owner, str)
    assert isinstance(repo_name, str)
    assert isinstance(sha, str)
    assert isinstance(github_access_token, str)
    assert repo_owner
    assert repo_name
    assert sha
    assert github_access_token

    github_url = 'https://api.github.com/repos/{repo_owner}/{repo_name}/commits/{sha}/status'.format(
        repo_owner=repo_owner, repo_name=repo_name, sha=sha
    )

    async with ClientSession(loop=loop) as session:
        headers = {
            'Authorization': 'token {token}'.format(token=github_access_token),
            'Content-Type': 'application/json'
        }
        async with session.get(github_url, headers=headers) as response:
            if response.status == 200:
                response_json = await response.json()
                try:
                    validated_response = github_status_validator.check(response_json)
                    if validated_response['sha'] == sha:
                        return validated_response['status']
                except trafaret.DataError:
                    pass


class BadgeApiView(web.View):
    """
    View that handles coverage badge
    """

    encoding = 'utf-8'
    content_type = 'image/svg+xml'
    public_token = None
    sha = None
    repo_id = None
    project = None
    commit = None

    async def get(self):
        self.public_token = self.request.match_info.get('token')
        if not self.public_token:
            raise HTTPNotFound

        self.repo_id, self.project = await get_repo(self.request.app.redis_pool, token=self.public_token)
        if self.project is None:
            raise HTTPNotFound

        coverage = self.project.get('coverage', 0)
        min_good_coverage = self.project.get('min_good_coverage', settings.MIN_GOOD_COVERAGE_PERCENTAGE)
        badge = render_badge(coverage, coverage >= min_good_coverage)

        return web.Response(
            body=badge,
            headers=MultiDict({'Content-Disposition': 'Inline'}),
            status=HTTP_OK,
            content_type=self.content_type
        )


class _CoverRageCommonApiView(web.View):
    """
    Common class for all Cover-Rage API views:
      - request data validation (`self.validator` should be set)
      - authorization (optional, set `self.check_authorization` to True)
      - searching repo in Redis
      - custom handling (should be overridden)
      - response / errors handling
    """

    validator = None
    raw_data = None
    public_token = None
    repo_id = None
    project = None
    validated_data = {}
    errors = {}
    encoding = 'utf-8'
    content_type = 'application/json'
    response_success_status = HTTP_OK
    response_error_status = HTTP_BAD_REQUEST
    check_authorization = False

    def __await__(self):  # pragma: no cover
        pass

    def __init__(self, request):
        super(_CoverRageCommonApiView, self).__init__(request)
        assert self.validator

    def auth(self):
        """
        Check that `Authorization` header is set with correct Cover-Rage private token.
        :raises: HTTPUnauthorized, HTTPForbidden
        """
        if 'Authorization' in self.request.headers:
            private_token = self.request.headers['Authorization']
        else:
            raise HTTPUnauthorized
        has_permission = (
            'rage_private_token' in self.project and
            self.project['rage_private_token'] == private_token
        )
        if not has_permission:
            raise HTTPForbidden

    def is_valid(self):
        try:
            self.validated_data = self.validator.check(self.raw_data)
            return True
        except trafaret.DataError:
            self.errors = trafaret.extract_error(self.validator, self.raw_data)
            return False

    async def handle(self):  # pragma: no cover
        raise NotImplemented

    async def post(self):
        """
        Common POST handler:
        --------------------
        1) Get project from public token.
        2) Check authorization (if `self.check_authorization` is True).
        3) Validate input data with `self.validator`.
        4) Do something with this data - override `handle` method.
        5) Return response.

        :raises: HTTPUnauthorized, HTTPForbidden, HTTPNotFound
        """

        # Get public token from url
        self.public_token = self.request.match_info.get('token')
        if not self.public_token:
            raise HTTPNotFound

        # Get repo id and data from Redis or raise 404
        self.repo_id, self.project = await get_repo(self.request.app.redis_pool, token=self.public_token)
        if self.project is None:
            raise HTTPNotFound

        # Check authorization if needed
        if self.check_authorization:
            self.auth()

        # Validate input data with `self.validator`
        self.raw_data = await self.request.json()
        if self.is_valid():
            response, response_status = await self.handle()
        else:
            response = deepcopy(self.errors)
            response_status = self.response_error_status

        # Return response
        return web.Response(
            body=bytes(json.dumps(response), self.encoding),
            status=response_status or self.response_success_status,
            content_type=self.content_type
        )


# TODO: Implement Bitbucket pull request events handling
class BitbucketPullRequestEventApiView(_CoverRageCommonApiView):

    pass


# TODO: Implement Bitbucket push events handling
class BitbucketPushEventApiView(_CoverRageCommonApiView):

    pass


class GithubPullRequestEventApiView(_CoverRageCommonApiView):
    """
    Handle Github `PULL REQUEST` event:
      - create / update mapping for commit (for actions: edited, opened, reopened / status: 200)
      - ignore other actions (status: 202)
    """

    validator = github_pull_request_event_validator

    async def handle(self):
        commit_data = github_pull_request_to_redis(self.validated_data)
        git_commit = commit_data.pop('sha')

        assert 'action' in commit_data
        if commit_data['action'] in GITHUB_PULL_REQUEST_TRIGGER_CREATE_UPDATE_ACTIONS:
            description = PENDING_CHECK_DESCRIPTION  # TODO: cover with tests
            status = GITHUB_PENDING_STATUS  # TODO: cover with tests
        elif commit_data['action'] in GITHUB_PULL_REQUEST_TRIGGER_REMOVE_ACTIONS:
            description = ERROR_CHECK_CLOSED_PULL_REQUEST_DESCRIPTION  # TODO: cover with tests
            status = GITHUB_ERROR_STATUS  # TODO: cover with tests
        else:
            # Ignore other Github pull request actions
            description = None
            status = None

        if all((description, status)):
            if not await update_commit(self.request.app.redis_pool, self.repo_id, git_commit, commit_data):
                return {'git_commit': 'Could not update commit'}, HTTP_BAD_REQUEST

            _, repo_owner, repo_name = split_repo_id(self.repo_id)
            if await send_github_status(
                self.request.app.loop,
                repo_owner,
                repo_name,
                git_commit,
                status,
                cover_rage_status_url(self.public_token, git_commit),
                description,
                COVER_RAGE_CONTEXT,
                self.project['repo_access_token']
            ):
                return {}, HTTP_OK
            else:
                return {'git_commit': 'Could not set status for commit'}, HTTP_BAD_REQUEST
        else:
            return {}, HTTP_ACCEPTED


class GithubPushEventApiView(_CoverRageCommonApiView):
    """
    Handle Github `PUSH` event - create mapping for commit.
    """

    validator = github_push_event_validator

    async def handle(self):
        commit_data = github_push_to_redis(self.validated_data)
        git_commit = commit_data.pop('sha')

        if not await update_commit(self.request.app.redis_pool, self.repo_id, git_commit, commit_data):
            return {'git_commit': 'Could not update commit'}, HTTP_BAD_REQUEST

        _, repo_owner, repo_name = split_repo_id(self.repo_id)
        if await send_github_status(
            self.request.app.loop,
            repo_owner,
            repo_name,
            git_commit,
            GITHUB_PENDING_STATUS,
            cover_rage_status_url(self.public_token, git_commit),
            PENDING_CHECK_DESCRIPTION,
            COVER_RAGE_CONTEXT,
            self.project['repo_access_token']
        ):
            return {}, HTTP_OK
        else:
            return {'git_commit': 'Could not set status for commit'}, HTTP_BAD_REQUEST


class ResultsApiView(_CoverRageCommonApiView):
    """
    Update repo information with coverage data from Cover-Rage client.
    """

    validator = cover_rage_client_validator
    check_authorization = True

    async def handle(self):
        coverage_data = cover_rage_client_to_redis(self.validated_data)
        git_commit = coverage_data.pop('git_commit')

        if not await update_coverage_for_commit(self.request.app.redis_pool, self.repo_id, git_commit, coverage_data):
            return {'git_commit': 'Could not update commit'}, HTTP_BAD_REQUEST

        current_coverage = int(round(self.validated_data['overall_coverage'] * 100))

        if await update_coverage_for_repo(self.request.app.redis_pool, self.repo_id, current_coverage):
            _, repo_owner, repo_name = split_repo_id(self.repo_id)
            if self.validated_data['uncovered_lines']:
                description = FAILURE_CHECK_DESCRIPTION  # TODO: cover with tests
                status = GITHUB_FAILURE_STATUS  # TODO: cover with tests
            else:
                description = SUCCESS_CHECK_DESCRIPTION  # TODO: cover with tests
                status = GITHUB_SUCCESS_STATUS  # TODO: cover with tests
            if await send_github_status(
                self.request.app.loop,
                repo_owner,
                repo_name,
                git_commit,
                status,
                cover_rage_status_url(self.public_token, git_commit),
                description,
                COVER_RAGE_CONTEXT,
                self.project['repo_access_token']
            ):
                return {}, HTTP_OK
            else:
                return {'git_commit': 'Could not set status for commit'}, HTTP_BAD_REQUEST
        else:
            return {'git_commit': 'Could not update repo'}, HTTP_BAD_REQUEST


class StatusApiView(web.View):
    """
    View that handles status of check for required commit
    """

    encoding = 'utf-8'
    content_type = 'application/json'
    public_token = None
    sha = None
    repo_id = None
    project = None
    commit = None

    async def get(self):
        """
        Get commit data from Redis
        :raises: HTTPNotFound
        :return:
        """
        self.public_token = self.request.match_info.get('token')
        if not self.public_token:
            raise HTTPNotFound

        self.sha = self.request.match_info.get('sha')
        if not self.sha:
            raise HTTPNotFound

        self.repo_id, self.project = await get_repo(self.request.app.redis_pool, token=self.public_token)
        if self.project is None:
            raise HTTPNotFound

        self.commit = await get_commit(self.request.app.redis_pool, self.repo_id, self.sha)
        if self.commit is None:
            raise HTTPNotFound

        return web.Response(
            body=bytes(json.dumps(self.commit), self.encoding),
            status=HTTP_OK,
            content_type=self.content_type
        )



#!/usr/bin/env python

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


setup(
    name='cover-rage-server',
    version='0.0.2',
    description='Utility (server) to check coverage for the last commit / pull request while running tests on CI.',
    long_description=open('README.md', 'r').read(),
    author='Alexander Ryabtsev',
    author_email='ryabtsev.alexander@gmail.com',
    url='https://github.com/alexryabtsev/cover_rage_server',
    packages=['cover_rage_server'],
    entry_points={
        'console_scripts': ['rage_cli=cover_rage_server.rage_cli:__main__'],
    },
    install_requires=open('requirements.txt', 'r').readlines(),
    license='MIT',
    classifiers=(
        b'Development Status :: 3 - Alpha',
        b'Environment :: Console',
        b'Intended Audience :: Developers',
        b'License :: OSI Approved :: MIT License',
        b'Natural Language :: English',
        b'Operating System :: OS Independent',
        b'Programming Language :: Python',
        b'Programming Language :: Python :: 3.5',
        b'Topic :: Software Development :: Testing',
    ),
)



from envparse import env


# Number of threads in executor
# https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
THREAD_POOL_SIZE = env.int('RAGE_SRV_THREAD_POOL_SIZE', default=4)


# Redis connection settings
REDIS_HOST = env.str('RAGE_SRV_REDIS_HOST', default='127.0.0.1')
REDIS_PORT = env.int('RAGE_SRV_REDIS_PORT', default=6379)
REDIS_DB = env.int('RAGE_SRV_REDIS_DB', default=1)


# Cover-Rage binding address settings. `SRV_SCHEME` could be one of: 'http', 'https'
SRV_SCHEME = env.str('RAGE_SRV_SCHEME', default='https')
SRV_HOST = env.str('RAGE_SRV_HOST')


# Cover-Rage API urls
SRV_API_SEND_RESULTS_URL = '/api/results/{public_token}/'
SRV_API_GITHUB_PUSH_EVENT_HOOK_URL = '/api/github/{public_token}/push/'
SRV_API_GITHUB_PULL_REQUEST_EVENT_HOOK_URL = '/api/github/{public_token}/pull_request/'

# TODO: change url from API end point to regular view
SRV_STATUS_URL = '/api/status/{public_token}/{sha}/'


# Minimal good coverage percentage (required for background color of badge)
# If project's coverage is less then value of this settings - badge background color will be red.
# If it's greater or equal - badge background color will be green.
# If it's zero (undefined) - badge background color will be orange.
MIN_GOOD_COVERAGE_PERCENTAGE = env.int('RAGE_SRV_MIN_GOOD_COVERAGE_PERCENTAGE', default=94)


# Almost all settings have default values. On the other hand - you're responsible for changing them.
# All the settings you specify will be used as is... except several values that could conflict with common sense:
assert THREAD_POOL_SIZE > 0
assert SRV_SCHEME in ('https', 'http',)
assert MIN_GOOD_COVERAGE_PERCENTAGE <= 100


# Settings without default values that you have to set:
assert SRV_HOST
